module chebfft

  implicit none

contains

! Subroutine Initch:
! setta i due vettori xch e chfns
! Nt = ordine della trasformata di Chebyshev
! xch = contiene le coordinate x dei punti di collocazione di Chebyshev
!       (dimensione Nt+1)
! chfns = contiene 4 funzioni utilizzate nelle sezioni di pre e post
!         processing della routine cheb (dimensione Nt+1, 4 )

  subroutine initch( xch, chfns, Nt )
    
    integer :: Nt
    double precision, intent(OUT) :: xch(:)
    double precision, intent(OUT) :: chfns(:,:)
    double precision :: tt, pin
    integer :: ntp1, ntp2, nt2, nt2p1, i

    ntp1 = Nt + 1
    ntp2 = Nt + 2
    nt2 = Nt / 2
    nt2p1 = Nt / 2 + 1
    pin = 4.d0 * datan( 1.d0 ) / dfloat( Nt )
    do i = 2, nt2
       xch( i ) = dcos( pin * dfloat( ntp1 - i ) )
       xch( ntp2 - i ) = -xch( i )
    enddo
    xch( 1 ) = -1.d0
    xch( ntp1 ) = +1.d0
    xch( nt2p1 ) = 0.d0
    do i = 2, Nt
       tt = dsqrt( 1.d0 - xch( i ) * xch( i ) )
       chfns( i, 1 ) = 0.5d0 * ( 1.d0 + xch( i ) + tt )
       chfns( i, 2 ) = 0.5d0 * ( 1.d0 - xch( i ) - tt )
       chfns( i, 3 ) = 0.25d0 * ( 1.0d0 + ( 1.0d0 - xch( i ) ) / tt )
       chfns( i, 4 ) = 0.25d0 * ( 1.0d0 - ( 1.0d0 - xch( i ) ) / tt )
    enddo
    chfns( 1, 1 ) = 0.d0
    chfns( ntp1, 1 ) = 1.0d0
    chfns( 1, 2 ) = 1.0d0
    chfns( ntp1, 2 ) = 0.d0
    chfns( ntp1, 3 ) = 0.25d0
    chfns( ntp1, 4 ) = 0.25d0
    
    return

  end subroutine initch


! Calcola la trasformata di Chebyshev diretta o inversa
! Nt = ordine della trasformata di Chebyshev
! f = funzione nello spazio fisico (dimensione Nt+1)
! fc = funzione nello spazio spettrale (dimensione Nt+1)
! wch = vettore di lavoro (dimensione Nt+1)
! chfns = vettore di lavoro inizializzato in initch (dimensione Nt+1, 4 )
! isg = direzione della trasformata: +1 -> trasformata diretta;
!                                    -1 -> trasformata inversa.

  subroutine cheb(f, fc, wch, chfns, Nt, isg)

    double precision, intent(INOUT), dimension ( : ) :: f, fc, wch
    double precision, intent(IN) :: chfns( : , : )
    integer :: Nt, isg

    double precision :: fc0, fn, f0
    integer :: ntp1, ntp2, in, isgn

    ntp1 = Nt + 1
    ntp2 = Nt + 2

    if ( isg == 1 ) then
       do in = 1, ntp1
          wch( in ) = chfns( in, 1 ) * f( in ) + chfns( in, 2 ) * f( ntp2 - in )
       enddo
       wch( ntp2 ) = 0.d0

       CALL REALFT(wch, Nt, 1)
       wch( ntp1 ) = wch( 2 )
       wch( 2 ) = 0.d0
       wch( ntp2 ) = 0.d0
       wch = wch/Nt
       do in = 4, Nt, 2
          wch( in ) = -wch( in )
       end do

       fc0 = 0.5d0 * ( f( 1 ) + f( ntp1 ) )
       do in = 1, Nt - 1
          fc0 = fc0 + f( in + 1 )
       enddo

       fc( 1 ) = fc0 / dfloat( Nt )
       fc( 2 ) = 2.d0 * ( wch( 1 ) - fc( 1 ) )

       do in = 3, ntp1, 2
          fc( in + 1 ) = fc( in - 1 ) - 4.d0 * wch( in + 1 )
       enddo

       do in = 2, Nt, 2
          fc( in + 1 ) = 2.d0 * ( wch( in + 1 ) - wch( in + 2 ) ) - fc( in + 2 )
       enddo
       fc( ntp2 ) = 0.d0

       return

! Trasformata inversa di Chebyshev

    else

       wch( 1 ) = 2.d0 * fc( 1 ) + fc( 2 )
       fc( Nt + 2 ) = 0.d0
       do in = 2, Nt, 2
          wch( in + 1 ) = fc( in + 1 ) + 0.5d0 * ( fc( in + 2 ) + fc( in ) )
          wch( in + 2 ) = 0.5d0 * ( fc( in ) - fc( in + 2 ) )
       enddo

       wch( 2 ) = wch( Nt + 1 )
       do in = 4, Nt, 2
          wch( in ) = -wch( in )
       end do
       CALL REALFT(wch, Nt, -1)
       wch = 2.d0*wch

       isgn = 1
       fn = fc( 1 ) + fc( Nt + 1 )
       f0 = fn
       do in = 1, Nt - 1
          isgn = - isgn
          fn = fn + fc( in + 1 )
          f0 = f0 + isgn * fc( in + 1 )
       enddo
       f( 1 ) = f0
       f( Nt + 1 ) = fn
       do in = 1, Nt - 1
          f( in + 1 ) = wch( in + 1 ) * chfns( in + 1, 3 ) + &
               wch( Nt - in + 1 ) * chfns( in + 1, 4 )
       enddo

      return

   endif

   end subroutine cheb

! Subroutines del numerical recipes per FFT

  SUBROUTINE realft(data,n,isign)
    IMPLICIT NONE
    
    INTEGER :: isign,n
    DOUBLE PRECISION :: data(n)
!   USES four1
    INTEGER :: i,i1,i2,i3,i4,n2p3
    DOUBLE PRECISION :: c1,c2,h1i,h1r,h2i,h2r,wis,wrs
    DOUBLE PRECISION :: theta,wi,wpi,wpr,wr,wtemp

    theta=3.141592653589793d0/dble(n/2)
    c1=0.5d0
    IF (isign == 1) THEN
       c2=-0.5d0
       CALL four1(data,n/2,+1)
    ELSE
       c2=0.5d0
       theta=-theta
    ENDIF
    wpr=-2.0d0*sin(0.5d0*theta)**2
    wpi=sin(theta)
    wr=1.0d0+wpr
    wi=wpi
    n2p3=n+3
    DO i=2,n/4
       i1=2*i-1
       i2=i1+1
       i3=n2p3-i2
       i4=i3+1
       wrs=dble(wr)
       wis=dble(wi)
       h1r=c1*(data(i1)+data(i3))
       h1i=c1*(data(i2)-data(i4))
       h2r=-c2*(data(i2)+data(i4))
       h2i=c2*(data(i1)-data(i3))
       data(i1)=h1r+wrs*h2r-wis*h2i
       data(i2)=h1i+wrs*h2i+wis*h2r
       data(i3)=h1r-wrs*h2r+wis*h2i
       data(i4)=-h1i+wrs*h2i+wis*h2r
       wtemp=wr
       wr=wr*wpr-wi*wpi+wr
       wi=wi*wpr+wtemp*wpi+wi
    ENDDO
    IF (isign == 1) THEN
       h1r=data(1)
       data(1)=h1r+data(2)
       data(2)=h1r-data(2)
    ELSE
       h1r=data(1)
       data(1)=c1*(h1r+data(2))
       data(2)=c1*(h1r-data(2))
       CALL four1(data,n/2,-1)
    ENDIF
  END SUBROUTINE realft




  SUBROUTINE four1(data,nn,isign)
    IMPLICIT NONE

    INTEGER :: isign,nn
    DOUBLE PRECISION :: data(2*nn)
    INTEGER :: i,istep,j,m,mmax,n
    DOUBLE PRECISION :: tempi,tempr
    DOUBLE PRECISION :: theta,wi,wpi,wpr,wr,wtemp
    
    n=2*nn
    j=1
    DO i=1,n,2
       IF (j > i) THEN
          tempr=data(j)
          tempi=data(j+1)
          data(j)=data(i)
          data(j+1)=data(i+1)
          data(i)=tempr
          data(i+1)=tempi
       ENDIF
       m=n/2
       DO WHILE ((m >= 2).AND.(j > m))
          j=j-m
          m=m/2
       ENDDO
       j=j+m
    ENDDO
    mmax=2
    DO WHILE (n.gt.mmax)
       istep=2*mmax
       theta=6.28318530717959d0/(isign*mmax)
       wpr=-2.d0*sin(0.5d0*theta)**2
       wpi=sin(theta)
       wr=1.d0
       wi=0.d0
       DO m=1,mmax,2
          DO i=m,n,istep
             j=i+mmax
             tempr=dble(wr)*data(j)-dble(wi)*data(j+1)
             tempi=dble(wr)*data(j+1)+dble(wi)*data(j)
             data(j)=data(i)-tempr
             data(j+1)=data(i+1)-tempi
             data(i)=data(i)+tempr
             data(i+1)=data(i+1)+tempi
          ENDDO
          wtemp=wr
          wr=wr*wpr-wi*wpi+wr
          wi=wi*wpr+wtemp*wpi+wi
       ENDDO
       mmax=istep
    ENDDO
  END SUBROUTINE four1

end module chebfft
